"""Python interface to OpenAi functionality"""

from oa.util import openai, grazed, djoin, app_data_dir
from oa.base import chat, complete, dalle, api
from oa.openai_specs import raw
from oa.tools import prompt_function, PromptFuncs
from oa import ask



"""Interface tools"""

from functools import partial
from typing import Optional, Callable
import string

from i2 import Sig
from oa.base import chat

# -----------------------------------------------------------------------------
# Helpers

string_formatter = string.Formatter()


def _extract_names_from_format_string(template: str):
    """Extract names from a string format template

    >>> _extract_names_from_format_string("Hello {name}! I am {bot_name}.")
    ('name', 'bot_name')

    """
    return tuple(
        name for _, name, _, _ in string_formatter.parse(template) if name is not None
    )


def _extract_defaults_from_format_string(template: str) -> dict:
    """Extract (name, specifier) from a string format template.

    >>> _extract_defaults_from_format_string(
    ...     "Hello {name}! I am {bot_name:chatGPT}."
    ... )
    {'bot_name': 'chatGPT'}

    """
    return dict(
        (name, specifier)
        for _, name, specifier, _ in string_formatter.parse(template)
        if name is not None and specifier != ""
    )


def _template_without_specifiers(template):
    """Uses remove any extras from a template string, leaving only text and fields.

    >>> _template_without_specifiers("A {normal}, {stra:nge} and an empty: {}.")
    'A {normal}, {stra} and an empty: {}.'

    """
    import string

    def gen():
        for text, field_name, *_ in string.Formatter().parse(template):
            text_ = text or ""
            if field_name is None:
                yield text_
            else:
                yield text_ + "{" + field_name + "}"

    return "".join(gen())


def string_format_embodier(template):
    names = _extract_names_from_format_string(template)
    sig = Sig(names).ch_kinds(**{name: Sig.KEYWORD_ONLY for name in names})

    @sig
    def templated_string_embodier(**kwargs):
        return template.format(**kwargs)

    return templated_string_embodier


# -----------------------------------------------------------------------------
# The meat


def prompt_function(
    template,
    *,
    defaults: Optional[dict] = None,
    template_to_names: Callable = _extract_names_from_format_string,
    template_to_defaults: Callable = _extract_defaults_from_format_string,
    embodier: Callable = string_format_embodier,
    arg_kinds: Optional[dict] = None,
    name=None,
    prompt_func=chat,
    prompt_func_kwargs=None,
    egress=None,
):
    """Convert a string template to a function that will call"""

    defaults = dict(template_to_defaults(template), **(defaults or {}))
    template = _template_without_specifiers(template)
    template_embodier = embodier(template)
    prompt_func_kwargs = prompt_func_kwargs or {}
    egress = egress or (lambda x: x)

    # TODO: Same logic replicated in string_format_embodier (what can we do?)
    names = template_to_names(template)
    # TODO: Too restrictive to require all names to be keyword only?
    arg_kinds = dict({name: Sig.KEYWORD_ONLY for name in names}, **(arg_kinds or {}))
    sig = Sig(names).ch_kinds(**arg_kinds)
    sig = sig.ch_defaults(
        _allow_reordering=True, **{name: default for name, default in defaults.items()}
    )

    @sig
    def ask_oa(*ask_oa_args, **ask_oa_kwargs):
        _kwargs = sig.kwargs_from_args_and_kwargs(
            ask_oa_args, ask_oa_kwargs, apply_defaults=True
        )
        # print(f"kwargs: {_kwargs}")
        __args, __kwargs = Sig(template_embodier).args_and_kwargs_from_kwargs(_kwargs)
        embodied_template = template_embodier(*__args, **__kwargs)
        # embodied_template = _call_forgivingly(
        #     template_embodier, args, kwargs #, enforce_sig=sig
        # )
        # embodied_template = call_forgivingly(template_embodier, *args, **kwargs)

        return egress(prompt_func(embodied_template, **prompt_func_kwargs))

    if name is not None:
        ask_oa.__name__ = name

    return ask_oa


from typing import Mapping, Optional, KT
from dol import TextFiles, filt_iter
from oa.util import templates_files
import os

chatgpt_templates_dir = os.path.join(templates_files, "chatgpt")
_ends_with_txt = lambda f: f.endswith(".txt")
chatgpt_templates = filt_iter(TextFiles(chatgpt_templates_dir), filt=_ends_with_txt)
dflt_function_key = lambda f: os.path.splitext(os.path.basename(f))[0]
dflt_factory_key = lambda f: os.path.splitext(os.path.basename(f))[-1]
_dflt_factories = {
    ".txt": prompt_function,
    "": prompt_function,
}
dflt_factories = tuple(_dflt_factories.items())

StoreKey = str
FuncName = str
FactoryKey = KT


class PromptFuncs:
    """Make AI enabled functions"""

    def __init__(
        self,
        template_store: Mapping = chatgpt_templates,
        *,
        function_key: Callable[[StoreKey], FuncName] = dflt_function_key,
        factory_key: Callable[[StoreKey], FactoryKey] = dflt_factory_key,
        factories: Callable[[FactoryKey], Callable] = dflt_factories,
        extra_template_kwargs: Optional[Mapping[StoreKey, Mapping]] = None,
    ):
        self._template_store = template_store
        self._function_key = function_key
        self._factory_key = factory_key
        self._factories = dict(factories)
        self._extra_template_kwargs = dict(extra_template_kwargs or {})
        self._functions = dict(self._mk_functions())
        self._inject_functions()

    def _mk_functions(self):
        for store_key, template in self._template_store.items():
            factory = self._factories[self._factory_key(store_key)]
            func_key = self._function_key(store_key)
            yield func_key, factory(
                template,
                name=func_key,
                **self._extra_template_kwargs.get(store_key, {}),
            )

    def _inject_functions(self):
        self.__dict__.update(self._functions)

    def __iter__(self):
        return iter(self._functions)

    def __getitem__(self, name):
        return self._functions[name]

    def __len__(self):
        return len(self._functions)



"""Base oa functionality"""
from itertools import chain
from functools import partial

from types import SimpleNamespace
from i2 import Sig, Param

from oa.util import openai, djoin
from oa.openai_specs import prompt_path, sig

api = None
# TODO: Expand this to include all the other endpoints (automatically?)
# api = SimpleNamespace(
#     chat=sig.CreateChatCompletionRequest(openai.ChatCompletion.create),
#     complete=sig.CreateCompletionRequest(openai.Completion.create),
#     dalle=sig.CreateImageRequest(openai.Image.create),
# )

prompt_dalle_path = partial(prompt_path, prefix=djoin('dalle'))
prompt_davinci_path = partial(prompt_path, prefix=djoin('davinci'))

# TODO: Understand the model/engine thing better and merge defaults if possible
DFLT_ENGINE = 'text-davinci-003'
DFLT_MODEL = 'gpt-3.5-turbo'

# TODO: Use oa.openai_specs sig to provide good signatures


def complete(prompt, model=None, **complete_params):
    if 'engine' in complete_params:
        model = complete_params.pop('engine')
    model = model or getattr(complete, 'engine', DFLT_ENGINE)
    text_resp = openai.Completion.create(model=model, prompt=prompt, **complete_params)
    return text_resp['choices'][0]['text']


complete.engine = DFLT_ENGINE


def _raw_chat(prompt=None, model=DFLT_MODEL, *, messages=None, **chat_params):
    if not ((prompt is None) ^ (messages is None)):
        raise ValueError('Either prompt or messages must be specified, but not both.')
    if prompt is not None:
        messages = [{'role': 'user', 'content': prompt}]
    return openai.ChatCompletion.create(messages=messages, model=model, **chat_params)


# chat_sig = sig.CreateChatCompletionRequest
# chat_sig = chat_sig.ch_defaults(model=DFLT_MODEL, messages=None)
# chat_sig = Sig([Param(name='prompt', default=None, annotation=str), *chat_sig.params])


def chat(prompt=None, model=DFLT_MODEL, *, messages=None, **chat_params):
    resp = _raw_chat(prompt=prompt, model=model, messages=messages, **chat_params)
    # TODO: Make attr and item getters more robust (use glom?)
    return resp['choices'][0].message.content


chat.raw = _raw_chat


def _raw_dalle(prompt, n=1, size='512x512', **image_create_params):
    return openai.Image.create(prompt=prompt, n=n, size=size, **image_create_params)


def dalle(prompt, n=1, size='512x512', **image_create_params):
    r = _raw_dalle(prompt=prompt, n=n, size=size, **image_create_params)
    return r['data'][0]['url']


def list_engine_ids():
    return [x['id'] for x in openai.Engine.list()['data']]



"""oa utils"""

from importlib.resources import files
import os
from config2py.tools import get_configs_local_store
import i2
import dol

import graze
from functools import partial
from config2py import get_config, ask_user_for_input
import openai  # pip install openai (see https://pypi.org/project/openai/)


def get_package_name():
    """Return current package name"""
    # return __name__.split('.')[0]
    # TODO: See if this works in all cases where module is in highest level of package
    #  but meanwhile, hardcode it:
    return 'oa'


# get app data dir path and ensure it exists
pkg_name = get_package_name()
data_files = files(pkg_name) / 'data'
templates_files = data_files / 'templates'
_root_app_data_dir = i2.get_app_data_folder()
app_data_dir = os.environ.get(
    f'{pkg_name.upper()}_APP_DATA_DIR', os.path.join(_root_app_data_dir, pkg_name),
)
app_data_dir = dol.ensure_dir(app_data_dir, verbose=f'Making app dir: {app_data_dir}')
djoin = partial(os.path.join, app_data_dir)

# _open_api_key_env_name = 'OPENAI_API_KEY'
# _api_key = os.environ.get(_open_api_key_env_name, None)
# if _api_key is None:
#     _api_key = getpass.getpass(
#         f"Please set your OpenAI API key and press enter to continue. "
#         f"I will put it in the environment variable {_open_api_key_env_name} "
#     )
# openai.api_key = _api_key


configs_local_store = get_configs_local_store(pkg_name)

openai.api_key = get_config(
    'OPENAI_API_KEY',
    sources=[
        # Try to find it in oa config
        configs_local_store,
        # Try to find it in os.environ (environmental variables)
        os.environ,
        # If not, ask the user to input it
        lambda k: ask_user_for_input(
            f'Please set your OpenAI API key and press enter to continue. '
            "If you don't have one, you can get one at "
            'https://platform.openai.com/account/api-keys. ',
            mask_input=True,
            masking_toggle_str='',
            egress=lambda v: configs_local_store.__setitem__(k, v),
        ),
    ],
)


_grazed_dir = dol.ensure_dir(os.path.join(app_data_dir, 'grazed'))
grazed = graze.Graze(rootdir=_grazed_dir)
